# -*- coding: utf-8 -*-
#
# modulemd_summarizer - Prints a summary of ModuleMD files
#
# Copyright © 2018, 2019 Red Hat, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
# Rafael dos Santos <rdossant@redhat.com>
# Nils Philippsen <nils@redhat.com>

import logging
import tempfile
from typing import Optional

from click import ClickException
import gi
gi.require_version('Modulemd', '1.0')  # noqa: E402
from gi.repository import Modulemd
import smartcols

from . import _fetchrepodata
from .config import runtime_config


log = logging.getLogger(__name__)


class ModuleProperty:

    properties_by_name = {}

    def __init__(self, name: str, property_location: str = None, display_name: str = None,
                 default_value=None):
        assert name not in self.properties_by_name

        self.properties_by_name[name] = self

        self.name = name
        self.property_location = property_location or f'metadata:{name}'
        self.display_name = display_name or name.replace("_", " ").title()
        self.default_value = default_value

    def __repr__(self):
        params = ['name']
        if self.property_location != f'metadata:{self.name}':
            params.append('property_location')
        if self.display_name != self.name.replace("_", " ").title():
            params.append('display_name')
        if self.default_value is not None:
            params.append('default_value')

        params_str = ", ".join(f"{p}={getattr(self, p)!r}" for p in params)

        return f"{type(self).__name__}({params_str})"

    def get_property_value(self,
                           module: Modulemd.Module,
                           module_defaults: Optional[Modulemd.Defaults]):
        objname, location = self.property_location.split(':')

        assert objname in ('metadata', 'buildinfo')
        assert location

        if objname == 'metadata':
            obj = module.props
        elif objname == 'buildinfo':
            obj = getattr(module, 'buildinfo', None)

        if not obj:
            return None

        value = obj

        for elem in location.split('.'):
            if not isinstance(value, dict):
                value = getattr(value, elem)
            else:
                value = value[elem]

        if value is None:
            return self.default_value

        return value

    def format_value(self,
                     module: Modulemd.Module,
                     module_defaults: Optional[Modulemd.Defaults]) -> str:
        value = self.get_property_value(module, module_defaults)
        return str(value) if value else "(n/a)"


class ModuleListProperty(ModuleProperty):

    def format_value(self,
                     module: Modulemd.Module,
                     module_defaults: Optional[Modulemd.Defaults]) -> str:
        value = self.get_property_value(module, module_defaults)

        return ", ".join(value) if value else ""


class ModuleStreamProperty(ModuleProperty):

    def get_property_value(self,
                           module: Modulemd.Module,
                           module_defaults: Optional[Modulemd.Defaults]):
        stream = super().get_property_value(module, module_defaults)

        if stream == module_defaults.props.default_stream:
            stream += " [d]"

        return stream


class ModuleProfilesProperty(ModuleListProperty):

    def get_property_value(self,
                           module: Modulemd.Module,
                           module_defaults: Optional[Modulemd.Defaults]):

        value = super().get_property_value(module, module_defaults)

        if not value:
            return

        profiles = sorted(value)

        if module_defaults:
            # convert SimpleSet values
            streams_dprofiles = {k: v.get()
                                 for k, v in module_defaults.props.profile_defaults.items()}

            for i, p in enumerate(profiles):
                if p in streams_dprofiles.get(module.props.stream, ()):
                    profiles[i] = p + " [d]"

        return profiles


class ModuleDependenciesProperty(ModuleListProperty):

    def get_property_value(self,
                           module: Modulemd.Module,
                           module_defaults: Optional[Modulemd.Defaults]):

        dependencies = super().get_property_value(module, module_defaults)

        return sorted(
            f"{dname}:{dstream}"
            for deps in dependencies
            for dname, stream_ss in sorted(deps.props.requires.items(),
                                           key=lambda x: x[0])
            for dstream in sorted(stream_ss.get())
        )


# declare properties shown by default
default_properties = [
    ModuleProperty('name'),
    ModuleStreamProperty('stream'),
    ModuleProperty('version'),
    # old module builds don't have the context set in their modulemd
    ModuleProperty('context', default_value="00000000"),
    ModuleProfilesProperty('profiles'),
    ModuleDependenciesProperty('dependencies'),
]

# declare optionally shown properties
optional_properties = [
    ModuleProperty('id', property_location='buildinfo:id', display_name="Build ID"),
    ModuleProperty('koji_tag', property_location='buildinfo:koji_tag'),
    ModuleProperty('owner', property_location='buildinfo:owner'),
    ModuleProperty('scmurl', property_location='buildinfo:scmurl', display_name="SCM URL"),
    ModuleProperty('time_completed', property_location='buildinfo:time_completed'),
    ModuleProperty('time_modified', property_location='buildinfo:time_modified'),
    ModuleProperty('time_submitted', property_location='buildinfo:time_submitted'),
]

all_properties = default_properties + optional_properties
all_property_names = [p.name for p in all_properties]


def _compute_prop_names(add_prop_names):
    prop_names = [p.name for p in default_properties]

    for n in add_prop_names or ():
        if n in prop_names:
            # don't show properties twice
            prop_names.remove(n)

        prop_names.append(n)

    return prop_names


def _print_summary(module_builds, module_defaults, props_shown, as_tree):
    tb = smartcols.Table()
    cols = {}

    for i, prop in enumerate(props_shown):
        cols[prop.name] = tb.new_column(prop.display_name)
        if i == 0:
            firstprop = prop
            firstcol = cols[prop.name]
            firstcol.tree = as_tree

    module_names_defaults = {d.props.module_name: d for d in module_defaults}

    sorted_builds = sorted(module_builds,
                           key=lambda mb: (mb.props.name, mb.props.stream,
                                           mb.props.version, mb.props.context))

    parent_ln = {}

    for mbuild in sorted_builds:
        module_name = mbuild.props.name
        module_defaults = module_names_defaults.get(
            module_name, Modulemd.Defaults(module_name=module_name))

        if as_tree:
            firstpropvalue = firstprop.format_value(mbuild, module_defaults)
            pl = parent_ln.get(firstpropvalue, None)
            if pl is None:
                pl = parent_ln.setdefault(firstpropvalue, tb.new_line())
            pl[firstcol] = firstpropvalue
        else:
            pl = None

        ln = tb.new_line(pl)

        for prop in props_shown:
            ln[cols[prop.name]] = prop.format_value(mbuild, module_defaults)

    if sorted_builds:
        print(tb)
        print('\nHint: [d]efault')
    else:
        print("No builds found!")


def _print_modulemds(module_builds):

    for m in module_builds:
        if hasattr(m, 'buildinfo'):
            print(m.buildinfo['modulemd'])
        else:
            print(m.dumps())


def summarize_modules(mfilter=None, yamls=None, urls=None, add_prop_names=None,
                      as_tree=False, action='summarize'):
    """
    Load Modulemd objects from each yaml file in the `yamls` list or each url
    in `urls` list and print a summary of the modules found with streams,
    context, version, profiles and dependencies information.
    If no files and no ulrs are passed, local cached metadata is used instead.

    *mfilter*: if present, restricts output to the nsvc supplied. Glob pattern
               can be used, e.g for filtering by stream: '*:master:*'

    *yamls*: list of yaml files to parse. If any, locally cached metadata will
             be ignored

    *urls*: list of repositories to read from. If any, locally cached metadata
            will be ignored.

    *add_prop_names*: list of additional properties to show

    *as_tree*: print the summary in tree format, grouping information.
    """

    assert action in ('summarize', 'modulemds')

    module_builds = []
    module_defaults = []

    if yamls or urls:
        if yamls:
            for yaml in yamls:
                try:
                    objects = Modulemd.objects_from_file(yaml)
                except gi.repository.GLib.GError as gerror:
                    raise ClickException(f"Could not read {yaml}: {gerror}") from gerror
        elif urls:
            for url in urls:
                with tempfile.TemporaryDirectory() as local_path:
                    rp = _fetchrepodata.RepoPaths(url, local_path)
                    _fetchrepodata._download_metadata_files(rp)
                    objects = _fetchrepodata._read_modules(rp)

        # distribute modules/builds and defaults into their own lists
        for o in objects:
            if isinstance(o, Modulemd.Module):
                module_builds.append(o)
            elif isinstance(o, Modulemd.Defaults):
                module_defaults.append(o)
            else:
                log.warning(f"Encountered unknown object in YAML: {o!r}")
    else:
        source = runtime_config['source']
        source_type = runtime_config['source_type']

        if source_type == 'buildsys':
            log.info("Querying build systems for information, this can take some time.")

        if mfilter:
            module_patterns = [p.split(":", 1)[0] for p in mfilter]
        else:
            module_patterns = None

        module_builds = list(source.enumerate_module_builds(nsvc_patterns=mfilter))
        module_defaults = list(source.enumerate_module_defaults(
            module_patterns=module_patterns
        ))

    if action == 'summarize':
        props_shown = [ModuleProperty.properties_by_name[n]
                       for n in _compute_prop_names(add_prop_names)]
        _print_summary(module_builds, module_defaults, props_shown, as_tree)
    elif action == 'modulemds':
        _print_modulemds(module_builds)
