# -*- coding: utf-8 -*-
#
# sources.connectors.pagure - get information about projects hosted on Pagure
#
# Copyright © 2018, 2019 Red Hat, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author:
# Nils Philippsen <nils@redhat.com>

import logging

from .rest import Consumer, paginate


log = logging.getLogger(__name__)


class PagureConsumer(Consumer):

    pagination = {
        'page_arg': 'page',
        'page_elem': 'pagination.page',
        'pages_elem': 'pagination.pages',
    }

    @paginate(data_elem='projects')
    async def get_projects(self, fork: bool = None, namespace: str = None, owner: str = None,
                           pattern: str = None, short: bool = None, tags: str = None,
                           username: str = None,
                           per_page: int = None, page: int = None) -> dict:

        """
        Lists projects on the Pagure instance.

        :param fork:        whether or not project forks should be listed
        :param namespace:   namespace beneath which the project lives
        :param owner:       filters by name of the project owner
        :param pattern:     glob pattern with which the project name must match
        :param short:       whether or not to shorten the project metadata
        :param tags:        list only projects with these tags
        :param username:    filters by users with commit rights
        :param per_page:    number of items returned per page (max 100)
        :param page:        which page to return

        :returns:           a JSON dictionary, returning the projects in the
                            'projects' element
        """
        log.debug(f"get_projects(fork={fork!r}, namespace={namespace!r},"
                  f" owner={owner!r}, pattern={pattern!r}, short={short!r},"
                  f" tags={tags!r}, username={username!r},"
                  f" per_page={per_page!r}, page={page!r})")
        params = {k: v for k, v in locals().items() if v is not None and k != 'self'}
        return await self.get_as_json('projects', params=params)

    async def get_project_branches(self, project, namespace=None):
        """
        Lists branches of a project on Pagure.

        :param project:     the project name
        :param namespace:   the project namespace (optional)

        :returns:           a JSON dictionary
        """

        log.debug(f"get_project_branches(project={project!r},"
                  f" namespace={namespace!r})")

        if namespace is None:
            return await self.get_as_json(f'{project}/git/branches')
        else:
            return await self.get_as_json(f'{namespace}/{project}/git/branches')
