# -*- coding: utf-8 -*-
#
# util.click - enhancements to the click module
#
# Copyright © 2019 Red Hat, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author:
# Nils Philippsen <nils@redhat.com>

import fnmatch

import click


class ChoiceGlob(click.Choice):
    """The choice-glob type allows a value to be checked as a glob pattern
    against a fixed set of supported values. Converts values to a list of
    matches.

    :param case_sensitive: Set to false to make choices case
        insensitive. Defaults to true.
    """

    name = 'choice-glob'

    def __repr__(self):
        return f"{type(self).__name__}({self.choices!r})"

    def convert(self, value, param, ctx):
        # try un-globbed value first
        try:
            return [super().convert(value, param, ctx)]
        except click.BadParameter:
            pass

        if self.case_sensitive:
            matchfn = fnmatch.fnmatchcase
        else:
            matchfn = fnmatch.fnmatch

        retvals = [c for c in self.choices if matchfn(c, value)]

        if not retvals:
            self.fail("invalid choice: {}. (must match any of {})".format(
                value, ", ".join(self.choices)))

        return retvals
